from __future__ import annotations

import dataclasses

import jax
from jax import lax
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, ArrayLike, Bool, Float, Key, PyTree

from lerax.policy import AbstractPolicyState

from .base_buffer import AbstractBuffer


class RolloutBuffer[StateType: AbstractPolicyState, ActType, ObsType](AbstractBuffer):
    observations: PyTree[ObsType]
    actions: PyTree[ActType]
    rewards: Float[Array, " *size"]
    dones: Bool[Array, " *size"]
    log_probs: Float[Array, " *size"]
    values: Float[Array, " *size"]
    returns: Float[Array, " *size"]
    advantages: Float[Array, " *size"]
    states: StateType

    def __init__(
        self,
        observations: PyTree[ObsType],
        actions: PyTree[ActType],
        rewards: Float[ArrayLike, " *size"],
        dones: Bool[ArrayLike, " *size"],
        log_probs: Float[ArrayLike, " *size"],
        values: Float[ArrayLike, " *size"],
        states: StateType,
        returns: Float[ArrayLike, " *size"] | None = None,
        advantages: Float[ArrayLike, " *size"] | None = None,
    ):
        self.observations = observations
        self.actions = actions
        self.rewards = jnp.asarray(rewards, dtype=float)
        self.dones = jnp.asarray(dones, dtype=bool)
        self.log_probs = jnp.asarray(log_probs, dtype=float)
        self.values = jnp.asarray(values, dtype=float)
        self.states = states
        self.returns = (
            jnp.asarray(returns, dtype=float)
            if returns is not None
            else jnp.full_like(values, jnp.nan, dtype=float)
        )
        self.advantages = (
            jnp.asarray(advantages, dtype=float)
            if advantages is not None
            else jnp.full_like(values, jnp.nan, dtype=float)
        )

    def compute_returns_and_advantages(
        self,
        last_value: Float[ArrayLike, ""],
        gae_lambda: Float[ArrayLike, ""],
        gamma: Float[ArrayLike, ""],
    ) -> RolloutBuffer[StateType, ActType, ObsType]:
        last_value = jnp.asarray(last_value)
        gamma = jnp.asarray(gamma)
        gae_lambda = jnp.asarray(gae_lambda)

        next_values = jnp.concatenate([self.values[1:], last_value[None]], axis=0)
        next_non_terminals = 1.0 - self.dones.astype(float)
        deltas = self.rewards + gamma * next_values * next_non_terminals - self.values
        discounts = gamma * gae_lambda * next_non_terminals

        def scan_fn(
            carry: Float[Array, ""], x: tuple[Float[Array, ""], Float[Array, ""]]
        ) -> tuple[Float[Array, ""], Float[Array, ""]]:
            delta, discount = x
            advantage = delta + discount * carry
            return advantage, advantage

        _, advantages = lax.scan(
            scan_fn, jnp.array(0.0), (deltas, discounts), reverse=True
        )
        returns = advantages + self.values

        return dataclasses.replace(self, advantages=advantages, returns=returns)

    def batches(
        self,
        batch_size: int,
        *,
        key: Key | None = None,
        batch_axes: tuple[int, ...] | int | None = None,
    ) -> RolloutBuffer[StateType, ActType, ObsType]:
        flat_self = self.flatten_axes(batch_axes)

        total = flat_self.rewards.shape[0]
        indices = jnp.arange(total) if key is None else jr.permutation(key, total)

        if total % batch_size != 0:
            total_trim = total - (total % batch_size)
            indices = indices[:total_trim]

        indices = indices.reshape(-1, batch_size)

        return jax.tree.map(lambda x: jnp.take(x, indices, axis=0), flat_self)

    def sample(
        self,
        batch_size: int,
        *,
        key: Key,
        batch_axes: tuple[int, ...] | int | None = None,
    ) -> RolloutBuffer[StateType, ActType, ObsType]:
        flat_self = self.flatten_axes(batch_axes)

        total = flat_self.rewards.shape[0]
        if batch_size > total:
            raise ValueError(
                f"Cannot sample batch_size={batch_size} from total={total} elements."
            )

        indices = jr.choice(key, total, shape=(batch_size,), replace=False)

        return jax.tree.map(lambda x: jnp.take(x, indices, axis=0), flat_self)

    @property
    def shape(self) -> tuple[int, ...]:
        return self.rewards.shape



from __future__ import annotations

from typing import ClassVar

import diffrax
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, ArrayLike, Bool, Float, Key

from lerax.render import (
    Abstract2DRenderer,
    AbstractRenderer,
    Color,
    PygameRenderer,
    Transform,
)
from lerax.space import Box

from .base_classic_control import (
    AbstractClassicControlEnv,
    AbstractClassicControlEnvState,
)


class PendulumState(AbstractClassicControlEnvState):
    y: Float[Array, "2"]
    t: Float[Array, ""]


class Pendulum(
    AbstractClassicControlEnv[PendulumState, Float[Array, ""], Float[Array, "3"]]
):
    name: ClassVar[str] = "Pendulum"

    action_space: Box
    observation_space: Box

    max_speed: Float[Array, ""]
    max_torque: Float[Array, ""]
    dt: Float[Array, ""]
    g: Float[Array, ""]
    m: Float[Array, ""]
    l: Float[Array, ""]

    solver: diffrax.AbstractSolver
    dt0: Float[Array, ""] | None
    stepsize_controller: diffrax.AbstractStepSizeController

    def __init__(
        self,
        *,
        max_speed: Float[ArrayLike, ""] = 8.0,
        max_torque: Float[ArrayLike, ""] = 2.0,
        g: Float[ArrayLike, ""] = 9.8,
        m: Float[ArrayLike, ""] = 1.0,
        l: Float[ArrayLike, ""] = 1.0,
        dt: Float[ArrayLike, ""] = 0.05,
        solver: diffrax.AbstractSolver | None = None,
        stepsize_controller: diffrax.AbstractStepSizeController | None = None,
    ):
        self.max_speed = jnp.array(max_speed)
        self.max_torque = jnp.array(max_torque)
        self.dt = jnp.array(dt)
        self.g = jnp.array(g)
        self.m = jnp.array(m)
        self.l = jnp.array(l)

        self.solver = solver or diffrax.Tsit5()
        is_adaptive = isinstance(self.solver, diffrax.AbstractAdaptiveSolver)
        self.dt0 = None if is_adaptive else self.dt
        if stepsize_controller is None:
            self.stepsize_controller = (
                diffrax.PIDController(rtol=1e-5, atol=1e-5)
                if is_adaptive
                else diffrax.ConstantStepSize()
            )
        else:
            self.stepsize_controller = stepsize_controller

        self.action_space = Box(-self.max_torque, self.max_torque)
        high = jnp.array([1.0, 1.0, self.max_speed])
        self.observation_space = Box(-high, high)

    def initial(self, *, key: Key) -> PendulumState:
        high = jnp.array([jnp.pi, 1.0])
        state = jr.uniform(key, shape=(2,), minval=-high, maxval=high)
        return PendulumState(y=state, t=jnp.array(0.0))

    def dynamics(
        self, t: Float[Array, ""], y: Float[Array, "2"], action: Float[Array, ""]
    ) -> Float[Array, "2"]:
        theta, theta_dot = y
        u = jnp.clip(action, -self.max_torque, self.max_torque)
        theta_dd = (
            3.0 * self.g / (2.0 * self.l) * jnp.sin(theta)
            + 3.0 / (self.m * self.l**2) * u
        )
        return jnp.array([theta_dot, theta_dd])

    def clip(self, y: Float[Array, "2"]) -> Float[Array, "2"]:
        theta, theta_dot = y
        theta = ((theta + jnp.pi) % (2 * jnp.pi)) - jnp.pi
        theta_dot = jnp.clip(theta_dot, -self.max_speed, self.max_speed)
        return jnp.array([theta, theta_dot])

    def observation(self, state: PendulumState, *, key: Key) -> Float[Array, "3"]:
        theta, theta_dot = state.y
        return jnp.array([jnp.cos(theta), jnp.sin(theta), theta_dot])

    def reward(
        self,
        state: PendulumState,
        action: Float[Array, ""],
        next_state: PendulumState,
        *,
        key: Key,
    ) -> Float[Array, ""]:
        theta, theta_dot = next_state.y
        u = jnp.clip(action, -self.max_torque, self.max_torque)
        cost = theta**2 + 0.1 * theta_dot**2 + 0.001 * (u**2)
        return -cost

    def terminal(self, state: PendulumState, *, key: Key) -> Bool[Array, ""]:
        return jnp.array(False)

    def render(self, state: PendulumState, renderer: AbstractRenderer):
        if not isinstance(renderer, Abstract2DRenderer):
            raise TypeError("Pendulum environment requires a 2D renderer.")

        theta = state.y[0]

        renderer.clear()

        rod_w = 0.2
        origin = jnp.array([0.0, 0.0])
        end = origin + self.l * jnp.array([jnp.sin(theta), jnp.cos(theta)])
        rod_color = Color(0.8, 0.3, 0.3)
        axle_color = Color(0.0, 0.0, 0.0)

        renderer.draw_circle(origin, radius=rod_w / 2, color=rod_color)
        renderer.draw_line(origin, end, color=rod_color, width=rod_w)
        renderer.draw_circle(end, radius=rod_w / 2, color=rod_color)
        renderer.draw_circle(origin, radius=0.05, color=axle_color)

        renderer.draw()

    def default_renderer(self) -> Abstract2DRenderer:
        width, height = 500, 500
        transform = Transform(
            scale=200.0,
            offset=jnp.array([width / 2, height / 2]),
            width=width,
            height=height,
            y_up=True,
        )
        return PygameRenderer(
            width=width,
            height=height,
            background_color=Color(1.0, 1.0, 1.0),
            transform=transform,
        )



from __future__ import annotations

import time
from abc import abstractmethod
from typing import Literal, Sequence

import equinox as eqx
from jax import lax
from jax import random as jr
from jaxtyping import Array, Bool, Float, Key

from lerax.render import AbstractRenderer
from lerax.space import AbstractSpace
from lerax.utils import unstack_pytree


class AbstractEnvLikeState(eqx.Module):

    @property
    @abstractmethod
    def unwrapped(self) -> AbstractEnvState:
        """Return the unwrapped environment state"""


class AbstractEnvLike[StateType: AbstractEnvLikeState, ActType, ObsType](eqx.Module):
    """Base class for RL environments or wrappers that behave like environments"""

    name: eqx.AbstractVar[str]

    action_space: eqx.AbstractVar[AbstractSpace[ActType]]
    observation_space: eqx.AbstractVar[AbstractSpace[ObsType]]

    @abstractmethod
    def initial(self, *, key: Key) -> StateType:
        """
        Generate the initial state of the environment.

        Args:
            key: A JAX PRNG key for any stochasticity in the initial state.

        Returns:
            An initial environment state.
        """

    @abstractmethod
    def transition(self, state: StateType, action: ActType, *, key: Key) -> StateType:
        """
        Update the environment state given an action.

        Args:
            state: The current environment state.
            action: The action to take.
            key: A JAX PRNG key for any stochasticity in the transition.

        Returns:
            The next environment state.
        """

    @abstractmethod
    def observation(self, state: StateType, *, key: Key) -> ObsType:
        """
        Generate an observation from the environment state.

        Args:
            state: The current environment state.
            key: A JAX PRNG key for any stochasticity in the observation.

        Returns:
            An observation corresponding to the environment state.
        """

    @abstractmethod
    def reward(
        self, state: StateType, action: ActType, next_state: StateType, *, key: Key
    ) -> Float[Array, ""]:
        """
        Generate a reward from the environment state transition.

        Args:
            state: The current environment state.
            action: The action taken.
            next_state: The next environment state.
            key: A JAX PRNG key for any stochasticity in the reward.

        Returns:
            A reward corresponding to the environment state transition.
        """

    @abstractmethod
    def terminal(self, state: StateType, *, key: Key) -> Bool[Array, ""]:
        """
        Determine whether the environment state is terminal.

        Args:
            state: The current environment state.
            key: A JAX PRNG key for any stochasticity in the terminal condition.

        Returns:
            A boolean indicating whether the environment state is terminal.
        """

    @abstractmethod
    def truncate(self, state: StateType) -> Bool[Array, ""]:
        """
        Determine whether the environment state is truncated.

        Args:
            state: The current environment state.

        Returns:
            A boolean indicating whether the environment state is truncated.
        """

    @abstractmethod
    def state_info(self, state: StateType) -> dict:
        """
        Generate additional info from the environment state.

        In many cases, this can simply return an empty dictionary.

        Args:
            state: The current environment state.

        Returns:
            A dictionary of additional info from the environment state.
        """

    @abstractmethod
    def transition_info(
        self, state: StateType, action: ActType, next_state: StateType
    ) -> dict:
        """
        Generate additional info from the environment state transition.

        In many cases, this can simply return an empty dictionary.

        Args:
            state: The current environment state.
            action: The action taken.
            next_state: The next environment state.

        Returns:
            A dictionary of additional info from the environment state transition.
        """

    @abstractmethod
    def default_renderer(self) -> AbstractRenderer:
        """
        Return the default renderer for the environment.

        Returns:
            An instance of AbstractRenderer for rendering the environment.
        """

    @abstractmethod
    def render(self, state: StateType, renderer: AbstractRenderer):
        """
        Render a frame from a state.

        Args:
            state: The environment state to render.
            renderer: The renderer to use for rendering.
        """

    def render_states(
        self,
        states: Sequence[StateType],
        renderer: AbstractRenderer | Literal["auto"] = "auto",
        dt: float = 0.0,
    ):
        """
        Render a sequence of frames from multiple states.

        Args:
            states: A sequence of environment states to render.
            renderer: The renderer to use for rendering. If "auto", uses the default renderer.
            dt: The time delay between rendering each frame, in seconds.
        """
        renderer = self.default_renderer() if renderer == "auto" else renderer
        renderer.open()
        for state in states:
            self.render(state, renderer)
            time.sleep(dt)
        renderer.close()

    def render_stacked(
        self,
        states: StateType,
        renderer: AbstractRenderer | Literal["auto"] = "auto",
        dt: float = 0.0,
    ):
        """
        Render multiple frames from stacked states.

        Stacked states are typically batched states stored in a pytree structure.

        Args:
            states: A pytree of stacked environment states to render.
            renderer: The renderer to use for rendering. If "auto", uses the default renderer.
            dt: The time delay between rendering each frame, in seconds.
        """
        self.render_states(unstack_pytree(states), renderer, dt)

    @property
    @abstractmethod
    def unwrapped(self) -> AbstractEnv:
        """
        Return the unwrapped environment.

        Returns:
            The unwrapped environment.
        """

    @eqx.filter_jit
    def reset(self, *, key: Key) -> tuple[StateType, ObsType, dict]:
        """
        Wrap the functional logic into a Gym API reset method.

        Args:
            key: A JAX PRNG key for any stochasticity in the reset.

        Returns:
            A tuple of the initial state, initial observation, and additional info.
        """
        initial_key, observation_key = jr.split(key, 2)
        state = self.initial(key=initial_key)
        observation = self.observation(state, key=observation_key)
        info = self.state_info(state)
        return state, observation, info

    @eqx.filter_jit
    def step(
        self, state: StateType, action: ActType, *, key: Key
    ) -> tuple[
        StateType, ObsType, Float[Array, ""], Bool[Array, ""], Bool[Array, ""], dict
    ]:
        """
        Wrap the functional logic into a Gym API step method.

        Args:
            state: The current environment state.
            action: The action to take.
            key: A JAX PRNG key for any stochasticity in the step.

        Returns:
            A tuple of the next state, observation, reward, terminal flag, truncate flag, and additional info.
        """
        transition_key, reward_key, terminal_key, reset_key = jr.split(key, 4)

        next_state = self.transition(state, action, key=transition_key)
        reward = self.reward(state, action, next_state, key=reward_key)
        terminal = self.terminal(next_state, key=terminal_key)
        truncate = self.truncate(next_state)
        info = self.transition_info(state, action, next_state)

        state = lax.cond(
            terminal | truncate, lambda: self.initial(key=reset_key), lambda: next_state
        )
        observation = self.observation(state, key=key)

        return state, observation, reward, terminal, truncate, info


class AbstractEnvState(AbstractEnvLikeState):
    """Base class for RL environment states."""

    @property
    def unwrapped(self) -> AbstractEnvState:
        """Return the unwrapped environment state"""
        return self


class AbstractEnv[StateType: AbstractEnvState, ActType, ObsType](
    AbstractEnvLike[StateType, ActType, ObsType]
):
    """
    Base class for RL environments.

    Attributes:
        name: The name of the environment
        action_space: The action space of the environment
        observation_space: The observation space of the environment
    """

    name: eqx.AbstractVar[str]

    action_space: eqx.AbstractVar[AbstractSpace[ActType]]
    observation_space: eqx.AbstractVar[AbstractSpace[ObsType]]

    @property
    def unwrapped(self) -> AbstractEnv[StateType, ActType, ObsType]:
        """Return the unwrapped environment"""
        return self



from __future__ import annotations

from distreqx import distributions
from jax import numpy as jnp
from jaxtyping import Array, ArrayLike, Float

from .base_distribution import AbstractDistreqxWrapper


class Normal(AbstractDistreqxWrapper[Float[Array, " dims"]]):
    """
    Normal distribution.

    Attributes:
        distribution: The underlying distreqx Normal distribution.

    Args:
        loc: The mean of the distribution.
        scale: The standard deviation of the distribution.
    """

    distribution: distributions.Normal

    def __init__(
        self,
        loc: Float[ArrayLike, " dims"],
        scale: Float[ArrayLike, " dims"],
    ):
        loc = jnp.asarray(loc)
        scale = jnp.asarray(scale)

        if loc.shape != scale.shape:
            raise ValueError("loc and scale must have the same shape.")

        self.distribution = distributions.Normal(loc=loc, scale=scale)

    @property
    def loc(self) -> Float[Array, " dims"]:
        return self.distribution.loc

    @property
    def scale(self) -> Float[Array, " dims"]:
        return self.distribution.scale



from __future__ import annotations

from abc import abstractmethod
from typing import Callable

import equinox as eqx
from jax import nn as jnn
from jax import numpy as jnp
from jaxtyping import Array, Float, Key, ScalarLike


class AbstractNODETerm(eqx.Module):
    """Base class for a neural ODE vector field term."""

    @abstractmethod
    def __call__(
        self, t: ScalarLike, z: Float[Array, " data_size"], args
    ) -> Float[Array, " data_size"]:
        """Vector field value."""


class MLPNODETerm(AbstractNODETerm):
    add_time: bool
    mlp: eqx.nn.MLP

    def __init__(
        self,
        data_size: int,
        width_size: int,
        depth: int,
        activation: Callable[[Array], Float[Array, " data_size"]] = jnn.softplus,
        final_activation: Callable[[Array], Float[Array, " data_size"]] = jnn.tanh,
        *,
        key: Key,
        add_time: bool = True,
    ):
        self.add_time = add_time
        self.mlp = eqx.nn.MLP(
            in_size=data_size + int(add_time),
            out_size=data_size,
            width_size=width_size,
            depth=depth,
            activation=activation,
            final_activation=final_activation,
            key=key,
        )

    def __call__(
        self, t: ScalarLike, z: Float[Array, " data_size"], args
    ) -> Float[Array, " data_size"]:
        if self.add_time:
            return self.mlp(jnp.concatenate([z, jnp.expand_dims(t, axis=-1)], axis=-1))
        else:
            return self.mlp(z)



from __future__ import annotations

from collections import OrderedDict
from typing import Any, Mapping

from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, Bool, Float, Key

from .base_space import AbstractSpace


class Dict(AbstractSpace[OrderedDict[str, Any]]):
    """
    A dictionary of spaces.

    Attributes:
        spaces: An ordered dictionary mapping keys to component spaces.

    Args:
        spaces: A mapping from keys to component spaces.
    """

    spaces: OrderedDict[str, AbstractSpace]

    def __init__(self, spaces: Mapping[str, AbstractSpace]):
        self.spaces = OrderedDict(spaces)

    @property
    def shape(self) -> None:
        return None

    def canonical(self) -> OrderedDict[str, Any]:
        return OrderedDict(
            {key: space.canonical() for key, space in self.spaces.items()}
        )

    def sample(self, key: Key) -> OrderedDict[str, Any]:
        return OrderedDict(
            {
                space_key: space.sample(subkey)
                for space_key, space, subkey in zip(
                    self.spaces.keys(),
                    self.spaces.values(),
                    jr.split(key, len(self.spaces)),
                )
            }
        )

    def contains(self, x: Any) -> Bool[Array, ""]:
        if not isinstance(x, OrderedDict):
            return jnp.array(False)

        if self.spaces.keys() != x.keys():
            return jnp.array(False)

        return jnp.array(
            [space.contains(x[key]) for key, space in self.spaces.items()]
        ).all()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, OrderedDict):
            return False

        return all(
            self_key == other_key and self_value == other_value
            for (self_key, self_value), (other_key, other_value) in zip(
                self.spaces.items(), other.items()
            )
        )

    def __repr__(self) -> str:
        return f"Dict({', '.join(f'{key}: {repr(space)}' for key, space in self.spaces.items())})"

    def __hash__(self) -> int:
        return hash(self.spaces.items())

    def flatten_sample(self, sample: OrderedDict[str, Any]) -> Float[Array, " size"]:
        parts = [
            space.flatten_sample(sample[key]) for key, space in self.spaces.items()
        ]
        return jnp.concatenate(parts)

    @property
    def flat_size(self) -> int:
        return sum(space.flat_size for space in self.spaces.values())

    def __getitem__(self, index: str) -> AbstractSpace:
        return self.spaces[index]

    def __len__(self) -> int:
        return len(self.spaces)



from __future__ import annotations

from distreqx import distributions
from jax import numpy as jnp
from jaxtyping import Array, ArrayLike, Float

from .base_distribution import AbstractDistreqxWrapper


class MultivariateNormalDiag(AbstractDistreqxWrapper[Float[Array, " dims"]]):
    """
    Multivariate Normal distribution with diagonal covariance.

    Attributes:
        distribution: The underlying distreqx MultivariateNormalDiag distribution.

    Args:
        loc: The mean of the distribution.
        scale_diag: The diagonal of the covariance matrix.
    """

    distribution: distributions.MultivariateNormalDiag

    def __init__(
        self,
        loc: Float[ArrayLike, " dims"] | None = None,
        scale_diag: Float[ArrayLike, " dims"] | None = None,
    ):
        loc = jnp.asarray(loc) if loc is not None else None
        scale_diag = jnp.asarray(scale_diag) if scale_diag is not None else None

        if (loc is not None and scale_diag is not None) and (
            loc.shape != scale_diag.shape
        ):
            raise ValueError("loc and scale_diag must have the same shape.")

        self.distribution = distributions.MultivariateNormalDiag(
            loc=loc, scale_diag=scale_diag
        )

    @property
    def loc(self) -> Float[Array, " dims"]:
        return self.distribution.loc

    @property
    def scale_diag(self) -> Float[Array, " dims"]:
        return self.distribution.scale_diag



from __future__ import annotations

from abc import abstractmethod

import equinox as eqx
import jax
import optax
from jax import lax
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, Int, Key, Scalar

from lerax.buffer import RolloutBuffer
from lerax.callback import (
    AbstractCallback,
    AbstractCallbackState,
    AbstractCallbackStepState,
    IterationContext,
    ResetContext,
    StepContext,
)
from lerax.env import AbstractEnvLike, AbstractEnvLikeState
from lerax.policy import (
    AbstractPolicyState,
    AbstractStatefulActorCriticPolicy,
    AbstractStatefulPolicy,
)
from lerax.space import Box
from lerax.utils import filter_scan

from .base_algorithm import AbstractAlgorithm, AbstractAlgorithmState, AbstractStepState


class OnPolicyStepState[PolicyType: AbstractStatefulPolicy](AbstractStepState):
    """
    State for on-policy algorithm steps.

    Attributes:
        env_state: The state of the environment.
        policy_state: The state of the policy.
        callback_state: The state of the callback for this step.
    """

    env_state: AbstractEnvLikeState
    policy_state: AbstractPolicyState
    callback_state: AbstractCallbackStepState

    @classmethod
    def initial(
        cls,
        env: AbstractEnvLike,
        policy: PolicyType,
        callback: AbstractCallback,
        key: Key,
    ) -> OnPolicyStepState[PolicyType]:
        """
        Initialize the step state for the on-policy algorithm.

        Resets the environment, policy, and callback states.

        Args:
            env: The environment to initialize.
            policy: The policy to initialize.
            callback: The callback to initialize.
            key: A JAX PRNG key.

        Returns:
            The initialized step state.
        """
        env_key, policy_key = jr.split(key, 2)
        env_state = env.initial(key=env_key)
        policy_state = policy.reset(key=policy_key)

        callback_states = callback.step_reset(ResetContext(locals()), key=key)

        return cls(env_state, policy_state, callback_states)


class OnPolicyState[PolicyType: AbstractStatefulPolicy](
    AbstractAlgorithmState[PolicyType]
):
    """
    State for on-policy algorithms.

    Attributes:
        iteration_count: The current iteration count.
        step_state: The state for the current step.
        env: The environment being used.
        policy: The policy being trained.
        opt_state: The optimizer state.
        callback_state: The state of the callback for this iteration.
    """

    iteration_count: Int[Array, ""]
    step_state: OnPolicyStepState[PolicyType]
    env: AbstractEnvLike
    policy: PolicyType
    opt_state: optax.OptState
    callback_state: AbstractCallbackState


class AbstractOnPolicyAlgorithm[PolicyType: AbstractStatefulActorCriticPolicy](
    AbstractAlgorithm[PolicyType, OnPolicyState[PolicyType]]
):
    """
    Base class for on-policy algorithms.

    Generates rollouts using the current policy and estimates advantages and
    returns using GAE. Trains the policy using the collected rollouts.

    Attributes:
        optimizer: The optimizer used for training the policy.
        gae_lambda: The GAE lambda parameter.
        gamma: The discount factor.
        num_envs: The number of parallel environments.
        num_steps: The number of steps to collect per environment.
        batch_size: The batch size for training.
    """

    optimizer: eqx.AbstractVar[optax.GradientTransformation]

    gae_lambda: eqx.AbstractVar[float]
    gamma: eqx.AbstractVar[float]

    num_envs: eqx.AbstractVar[int]
    num_steps: eqx.AbstractVar[int]
    batch_size: eqx.AbstractVar[int]

    def num_iterations(self, total_timesteps: int) -> int:
        return total_timesteps // (self.num_envs * self.num_steps)

    @abstractmethod
    def per_step(
        self, step_state: OnPolicyStepState[PolicyType]
    ) -> OnPolicyStepState[PolicyType]:
        """Process the step carry after each step."""

    def step(
        self,
        env: AbstractEnvLike,
        policy: PolicyType,
        state: OnPolicyStepState[PolicyType],
        *,
        key: Key,
        callback: AbstractCallback,
    ) -> tuple[OnPolicyStepState[PolicyType], RolloutBuffer]:
        (
            action_key,
            transition_key,
            observation_key,
            reward_key,
            terminal_key,
            bootstrap_key,
            env_reset_key,
            policy_reset_key,
            callback_key,
        ) = jr.split(key, 9)

        observation = env.observation(state.env_state, key=observation_key)
        next_policy_state, action, value, log_prob = policy.action_and_value(
            state.policy_state, observation, key=action_key
        )

        if isinstance(env.action_space, Box):
            clipped_action = jnp.clip(
                action,
                env.action_space.low,
                env.action_space.high,
            )
        else:
            clipped_action = action

        next_env_state = env.transition(
            state.env_state, clipped_action, key=transition_key
        )

        reward = env.reward(
            state.env_state, clipped_action, next_env_state, key=reward_key
        )
        termination = env.terminal(next_env_state, key=terminal_key)
        truncation = env.truncate(next_env_state)
        done = termination | truncation

        # Bootstrap reward if truncated
        # TODO: Check if a non-branched approach is faster
        bootstrapped_reward = lax.cond(
            truncation,
            lambda: reward
            + self.gamma
            * policy.value(
                next_policy_state, env.observation(next_env_state, key=bootstrap_key)
            )[1],
            lambda: reward,
        )

        # Reset environment if done
        next_env_state = lax.cond(
            done, lambda: env.initial(key=env_reset_key), lambda: next_env_state
        )

        # Reset policy state if done
        next_policy_state = lax.cond(
            done, lambda: policy.reset(key=policy_reset_key), lambda: next_policy_state
        )

        callback_state = callback.on_step(
            StepContext(
                state.callback_state, env, policy, done, bootstrapped_reward, locals()
            ),
            key=callback_key,
        )

        return (
            OnPolicyStepState(next_env_state, next_policy_state, callback_state),
            RolloutBuffer(
                observations=observation,
                actions=clipped_action,
                rewards=bootstrapped_reward,
                dones=done,
                log_probs=log_prob,
                values=value,
                states=state.policy_state,
            ),
        )

    def collect_rollout(
        self,
        env: AbstractEnvLike,
        policy: PolicyType,
        step_state: OnPolicyStepState[PolicyType],
        callback: AbstractCallback,
        key: Key,
    ) -> tuple[OnPolicyStepState[PolicyType], RolloutBuffer]:
        """Collect a rollout using the current policy."""
        key, observation_key = jr.split(key, 2)

        def scan_step(
            carry: OnPolicyStepState[PolicyType], key: Key
        ) -> tuple[OnPolicyStepState[PolicyType], RolloutBuffer]:
            carry, rollout = self.step(
                env,
                policy,
                carry,
                key=key,
                callback=callback,
            )
            return self.per_step(carry), rollout

        step_state, rollout_buffer = filter_scan(
            scan_step, step_state, jr.split(key, self.num_steps)
        )

        _, next_value = policy.value(
            step_state.policy_state,
            env.observation(step_state.env_state, key=observation_key),
        )
        rollout_buffer = rollout_buffer.compute_returns_and_advantages(
            next_value, self.gae_lambda, self.gamma
        )
        return step_state, rollout_buffer

    @abstractmethod
    def train(
        self,
        policy: PolicyType,
        opt_state: optax.OptState,
        buffer: RolloutBuffer,
        *,
        key: Key,
    ) -> tuple[PolicyType, optax.OptState, dict[str, Scalar]]:
        """
        Train the policy using the rollout buffer.

        Args:
            policy: The current policy.
            opt_state: The current optimizer state.
            buffer: The rollout buffer containing collected experiences.
            key: A JAX PRNG key.

        Returns:
            A tuple containing the updated policy, updated optimizer state,
            and a log dictionary.
        """

    def reset(
        self,
        env: AbstractEnvLike,
        policy: PolicyType,
        *,
        key: Key,
        callback: AbstractCallback,
    ) -> OnPolicyState[PolicyType]:
        step_key, callback_key = jr.split(key, 2)

        if self.num_envs == 1:
            step_state = OnPolicyStepState.initial(env, policy, callback, step_key)
        else:
            step_state = jax.vmap(
                OnPolicyStepState.initial, in_axes=(None, None, None, 0)
            )(env, policy, callback, jr.split(step_key, self.num_envs))

        callback_state = callback.reset(ResetContext(locals()), key=callback_key)

        return OnPolicyState(
            jnp.array(0, dtype=int),
            step_state,
            env,
            policy,
            self.optimizer.init(eqx.filter(policy, eqx.is_inexact_array)),
            callback_state,
        )

    def iteration(
        self,
        state: OnPolicyState[PolicyType],
        *,
        key: Key,
        callback: AbstractCallback,
    ) -> OnPolicyState[PolicyType]:
        rollout_key, train_key, callback_key = jr.split(key, 3)

        if self.num_envs == 1:
            step_state, rollout_buffer = self.collect_rollout(
                state.env,
                state.policy,
                state.step_state,
                callback,
                rollout_key,
            )
        else:
            step_state, rollout_buffer = eqx.filter_vmap(
                self.collect_rollout, in_axes=(None, None, 0, None, 0)
            )(
                state.env,
                state.policy,
                state.step_state,
                callback,
                jr.split(rollout_key, self.num_envs),
            )

        policy, opt_state, log = self.train(
            state.policy, state.opt_state, rollout_buffer, key=train_key
        )

        state = state.next(step_state, policy, opt_state)

        state = state.with_callback_states(
            callback.on_iteration(
                IterationContext(
                    state.callback_state,
                    state.step_state.callback_state,
                    state.env,
                    state.policy,
                    state.iteration_count,
                    state.opt_state,
                    log,
                    locals(),
                ),
                key=callback_key,
            )
        )

        return self.per_iteration(state)



from __future__ import annotations

from dataclasses import replace

import equinox as eqx
import jax

try:
    import mujoco
    from mujoco import mjx
except ImportError as e:
    raise ImportError(
        "MuJoCo environments require the optional mujoco dependency. "
        "Install it with: pip install 'lerax[mujoco]'"
    ) from e

from jax import lax
from jax import numpy as jnp
from jaxtyping import Array, Bool, Float, Key, PyTree

from lerax.render import AbstractRenderer
from lerax.render.mujoco_renderer import MujocoRenderer
from lerax.space import Box

from ..base_env import AbstractEnv, AbstractEnvState


def fix_dtype(x: Array) -> Array:
    mapping = {
        jnp.floating: jnp.float_,
        jnp.integer: jnp.int_,
        jnp.uint: jnp.uint,
        jnp.bool: jnp.bool_,
    }

    for base_type, target_type in mapping.items():
        if jnp.issubdtype(x.dtype, base_type):
            return x.astype(target_type)

    return x


def tree_fix_dtype(tree: PyTree) -> PyTree:
    return jax.tree.map(fix_dtype, tree)


class MujocoEnvState(AbstractEnvState):

    sim_state: eqx.AbstractVar[mjx.Data] = eqx.field(converter=tree_fix_dtype)
    t: eqx.AbstractVar[Float[Array, ""]]


class AbstractMujocoEnv[
    ActType: Float[Array, "..."],
    ObsType: Float[Array, "..."],
](AbstractEnv[MujocoEnvState, ActType, ObsType]):
    name: eqx.AbstractVar[str]

    action_space: eqx.AbstractVar[Box]
    observation_space: eqx.AbstractVar[Box]

    model: eqx.AbstractVar[mjx.Model]
    mujoco_model: eqx.AbstractVar[mujoco.MjModel]
    frame_skip: eqx.AbstractVar[int]
    dt: eqx.AbstractVar[float]

    def transition(
        self, state: MujocoEnvState, action: ActType, *, key: Key
    ) -> MujocoEnvState:
        def step_once(data: mjx.Data, _) -> tuple[mjx.Data, None]:
            data = mjx.step(self.model, data)
            return data, None

        data = state.sim_state.replace(ctrl=action)
        data, _ = lax.scan(step_once, data, None, length=self.frame_skip)

        return replace(state, sim_state=data, t=state.t + self.dt)

    def truncate(self, state: MujocoEnvState) -> Bool[Array, ""]:
        return jnp.array(False)

    def state_info(self, state: MujocoEnvState) -> dict:
        return {}

    def default_renderer(self) -> MujocoRenderer:
        return MujocoRenderer(self.mujoco_model)

    def render(self, state: MujocoEnvState, renderer: AbstractRenderer):
        if not isinstance(renderer, MujocoRenderer):
            raise TypeError("Mujoco environment requires a Mujoco renderer.")

        renderer.render(state.sim_state)
        renderer.draw()
